﻿/**** **** **** ****/ /**** **** **** ****/ /**** **** **** ****/ /**** **** **** ****/
/**** **** **** ****/ /**** **** **** ****/ /**** **** **** ****/ /**** **** **** ****/
/**** **** **** ****/ /**** **** **** ****/ /**** **** **** ****/ /**** **** **** ****/
/**** **** **** ****/ /**** **** **** ****/ /**** **** **** ****/ /**** **** **** ****/
/**/
/* este programa es capaz de generar la enésima permutación de cualquier lista */

/*genera la eNésima permutación según una matriz */
generar_permutacion( Momento, Matriz, Matriz_permutada):-

	longitud( N, Matriz),    % calcula la longitud N de la lista Matriz.

	lista_fantasma( Matriz_permutada, N),   % genera una lista Matriz_permutada, no instaciada, de longitud N.

	instanciar_matriz( Momento, Matriz, N, Matriz_permutada).     %instancia la lista Matriz_permutada

/*descripción de las variables:

Momento: es el número de orden de una permutación específica.
Por ejemplo si Momento es 4 se está refiriendo a la cuarta permutación.
Es una variable de entrada, o sea -Momento. Debe estar instanciada.

Matriz: Es la lista de la que se desea obtener una permutación eNésima
específica. Por ejemplo Matriz puede ser [ t, d, z, s]
Es una variable de entrada, o sea -Matriz. Debe estar instanciada.

Matriz_permutada: es la lista que contiene la permutación eNésima de
Matriz. En este caso su valor sería [t, s, d, z]
Es una variable de salida, o sea +Matriz_permutada. No debe estar
instanciada.

En resumen:

1 ?- generar_permutacion( 4, [t,d,z,s], Matriz_permutada).
Matriz_permutada = [t, s, d, z]

2 ?- generar_permutacion( 9, [t,d,z,s], Matriz_permutada).
Matriz_permutada = [z, t, d, s] .

3 ?- generar_permutacion( 12, [t,d,z,s], Matriz_permutada).
Matriz_permutada = [s, t, z, d] .

4 ?- generar_permutacion( 24, [t,d,z,s], Matriz_permutada).
Matriz_permutada = [s, z, d, t] .


*/


/*instanciar una lista con una permutación eNésima*/
instanciar_matriz( _, _, 0, _):-
	!,
	true.
instanciar_matriz( Momento, [ Cabeza| Cola], N, Matriz_permutada):-
	calcular_posicion( Momento, N, Posicion, Segmento_divisor),
	instanciar_matriz_posicion( Posicion, Cabeza, N, Matriz_permutada),
	calcular_parametros( N, Momento, Posicion, Segmento_divisor,
		     N_auxiliar, Momento_auxiliar),
	instanciar_matriz( Momento_auxiliar, Cola, N_auxiliar, Matriz_permutada).

/* calcular posicion según momento aleatoreo*/
calcular_posicion( Momento, N, Posicion, Segmento_divisor):-
	factorial( N, Factorial),
	Segmento_divisor is Factorial / N,
	Posicion_auxiliar is Momento // Segmento_divisor,
	Resto is Momento mod Segmento_divisor,
	corregir_posicion( Posicion_auxiliar, Resto, Posicion).

/* La sintaxis es factorial(N, F) -> Factorial de N es F (el resultado se guarda en F)
el factorial se llama recursivamente dejando el resultado en F */
factorial(0, 1) :- !.
factorial(N, F) :- N1 is N - 1, factorial(N1, F1), F is N * F1.

/* corrige posicion segun resto*/
corregir_posicion( Posicion, 0, Posicion):-
	!,
	true.
corregir_posicion( Posicion_auxiliar, _, Posicion):-
	Posicion is Posicion_auxiliar + 1.

/*instancia el eNésimo espacio vacío de una lista*/
instanciar_matriz_posicion( 1, Elemento, _, [ Cabeza| _]):-
	var( Cabeza),
	Cabeza = Elemento. % ya se puede introducir un elemento de cualquier naturaleza
instanciar_matriz_posicion( Posicion, Elemento, N, [ Cabeza| Cola]):-
	var( Cabeza),
	Posicion_auxiliar is Posicion - 1,
	instanciar_matriz_posicion( Posicion_auxiliar, Elemento, N, Cola).
instanciar_matriz_posicion( Posicion, Elemento, N, [ Cabeza| Cola]):-
	nonvar( Cabeza),
	instanciar_matriz_posicion( Posicion, Elemento, N, Cola).

/* calcular parámetros recurrentes */
calcular_parametros( N, Momento_anterior, Posicion, Segmento_divisor,
		     M, Momento_posterior):-
	M is N - 1,
	Momento_posterior is
	Momento_anterior - ( Posicion - 1) * Segmento_divisor.

/************* Lista Fantasma ********************/
/**** genera una lista de N elementos no instanciados.
 Se deriva del predicado "Permutaci¢n"   */
lista_fantasma( [], 0):-!.
lista_fantasma( [ _| Cola_fantasma], N):-
	N_aux is N - 1,
	lista_fantasma(  Cola_fantasma, N_aux).

/*** Longitud de una lista ***/

longitud( 0, []):- !.
longitud( N, [ _| Cola]):-
	longitud( N_aux, Cola),
	N is N_aux + 1.

